import abc
import asyncio
import time
from typing import Any, Iterable

from imap_tools.utils import EmailAddress
from langchain.chat_models import init_chat_model
from langchain_core.messages import HumanMessage, SystemMessage
from loguru import logger

from llmailbot.config import ModelSpec, ReplyConfig
from llmailbot.email.model import IMAPMessage, IMAPRawMessage, SimpleEmailMessage
from llmailbot.queue import AsyncQueue
from llmailbot.taskrun import AsyncTask, TaskDone


def quoted(txt: str) -> str:
    return "> " + "\n> ".join(txt.splitlines())


def quote_email(email: IMAPMessage) -> str:
    quote_title = f"{email.addr_from.name or email.addr_from.email} said"
    if email.date:
        quote_title += f" at {email.date.strftime('%Y-%m-%d %H:%M')}:"
    return f"{quote_title}\n\n{quoted(email.text)}"


class MailBot(abc.ABC):
    """
    MailBot is the interface class for mailbots.
    """

    @abc.abstractmethod
    async def compose_reply(
        self,
        spec: ModelSpec,
        bot_email: str,
        user_email: str,
        conversation: str,
    ) -> str:
        pass

    def __init__(self, specs: list[ModelSpec]):
        self.specs = specs

    def _get_spec(self, to_addrs: list[str]) -> tuple[str, ModelSpec] | tuple[None, None]:
        for addr in to_addrs:
            for spec in self.specs:
                if spec._address_regex is not None:
                    if spec._address_regex.match(addr):
                        return addr, spec
                else:
                    if spec.address == addr:
                        return addr, spec

        logger.warning("Received email for address with no matching configuration: {}", to_addrs)
        return None, None

    async def reply(self, email: IMAPMessage) -> SimpleEmailMessage | None:
        bot_email, spec = self._get_spec([a.email for a in email.addrs_to])
        if spec is None or bot_email is None:
            return None
        conversation = str(email)
        reply_body = await self.compose_reply(spec, bot_email, email.addr_from.email, conversation)
        reply_body = reply_body + "\n\n" + quote_email(email)
        return email.create_reply(EmailAddress(spec.name, bot_email), reply_body)


class HelloMailBot(MailBot):
    """
    HelloMailBot is a test MailBot implementation that generates a simple hardcoded reply.
    """

    async def compose_reply(
        self,
        spec: ModelSpec,
        bot_email: str,
        user_email: str,
        conversation: str,
    ) -> str:
        await asyncio.sleep(3)
        return f"Hello! My name is {spec.name}. I'm not very smart yet."


DEFAULT_CONFIGURABLE_FIELDS = frozenset(
    {
        "model",
        "model_provider",
        "max_tokens",
        "temperature",
        "base_url",
    }
)


class LangChainMailBot(MailBot):
    """
    LangChainMailBot is a MailBot implementation that generates LLM replies,
    using langchain chat models.

    It can reply on behalf of any number of "bots" by choosing a configuration
    based on the email To header.
    """

    def __init__(
        self,
        specs: list[ModelSpec],
        configurable_fields: Iterable[str] | None = None,
    ):
        super().__init__(specs)
        self.configurable_fields = set(configurable_fields or DEFAULT_CONFIGURABLE_FIELDS)
        self.chat_model = init_chat_model(
            max_retries=3,
            timeout=10,
            configurable_fields=list(self.configurable_fields),
        )

    def _build_system_prompt(self, spec: ModelSpec, bot_email: str, user_email: str) -> str:
        return spec.system_prompt.format(
            name=spec.name,
            bot_email=bot_email,
            user_email=user_email,
        )

    def _get_chat_model_config(self, spec: ModelSpec, bot_email: str) -> dict[str, Any]:
        model_config = {}
        for k, v in spec.chat_model_config(bot_email).items():
            k = k.lower()
            if k in self.configurable_fields:
                model_config[k] = v
            else:
                logger.warning("Model config field {} was provided but is not configurable", k)

        return model_config

    async def compose_reply(
        self,
        spec: ModelSpec,
        bot_email: str,
        user_email: str,
        conversation: str,
    ) -> str:
        system_prompt = self._build_system_prompt(spec, bot_email, user_email)
        model_config = self._get_chat_model_config(spec, bot_email)
        truncated_msg = conversation[: spec.max_input_length]
        logger.debug(
            "Composing reply with config={}, prompt={}, message length={}",
            model_config,
            system_prompt,
            len(truncated_msg),
        )
        messages = [SystemMessage(system_prompt), HumanMessage(truncated_msg)]
        try:
            stime = time.time()
            chat_model = self.chat_model.with_config(**model_config)
            response = await chat_model.ainvoke(messages)
            logger.debug(
                "{} {} response time: {}",
                spec.name,
                model_config["model"],
                time.time() - stime,
            )
            if isinstance(response.content, str):
                return response.content
            else:
                raise ValueError(f"Expected a str response, got: {repr(response.content)}")
        except Exception as e:
            logger.error("Chat model error: {}", e)
            return "We apologize, there was a system error. Your email is lost forever :("


def make_mailbot(
    specs: list[ModelSpec],
    configurable_fields: Iterable[str] | None = None,
    hello_bot: bool = False,
) -> MailBot:
    if hello_bot:
        return HelloMailBot(specs)
    else:
        return LangChainMailBot(specs, configurable_fields)


class BotReplyTask(AsyncTask[None]):
    """
    BotReplyTask is a task that runs a mailbot to reply to a single email.

    Args:
        mailbot: The mailbot instance to run
        email: The user email being replied to
        send_queue: The outgoing email queue
        retries: Number of retries on failure
        queue_timeout: Timeout for queue operations in seconds (high value hangs on exit)
    """

    def __init__(
        self,
        mailbot: MailBot,
        email: IMAPMessage,
        send_queue: AsyncQueue[SimpleEmailMessage],
        retries: int = 3,
    ):
        super().__init__(f"{self.__class__.__name__}<{mailbot.__class__.__name__}>")
        self.name = self._name
        self.mailbot = mailbot
        self.email = email
        self.sendq = send_queue
        self.retries = retries

    async def qput(self, message: SimpleEmailMessage) -> None:
        logger.trace(
            "Putting reply in mail queue",
        )
        await self.sendq.put(message)

    async def run(self) -> TaskDone[None] | None:
        email = self.email
        # TODO: implement retry logic in TaskRunner
        for retry_num in range(self.retries + 1):
            try:
                logger.info(
                    "{} generating reply to {}",
                    self.name,
                    email.summary(),
                )
                reply = await self.mailbot.reply(email)
                if reply is not None:
                    await self.qput(reply)
                break
            except Exception:
                logger.exception(
                    "Exception replying to email {} (retry {} of {})",
                    email.summary(),
                    retry_num,
                    self.retries,
                )
        return TaskDone(None)


class BotReplySpawnTask(AsyncTask[None]):
    """
    BotReplySpawnTask creates and runs a BotReplyTask for each email
    received from the incoming email queue.

    Args:
        mailbot: The mailbot instance to run
        recv_queue: Incoming email queue
        send_queue: Outgoing email queue
        retries: Number of retries on failure for reply task (default: 3)
        one_at_a_time: If True, will only spawn one reply task at a time (default: False)
        instance_n: Instance number for logging (default: None)
    """

    def __init__(
        self,
        mailbot: MailBot,
        recv_queue: AsyncQueue[IMAPRawMessage],
        send_queue: AsyncQueue[SimpleEmailMessage],
        retries: int = 3,
        one_at_a_time: bool = False,
        instance_n: int | None = None,
    ):
        self.mailbot = mailbot
        self.recvq = recv_queue
        self.sendq = send_queue
        self.retries = retries
        self.one_at_a_time = one_at_a_time
        self.name = f"{self.__class__.__name__}<{mailbot.__class__.__name__}>"
        if instance_n is not None:
            self.name += f".{instance_n}"
        super().__init__(self.name)

    async def run(self) -> TaskDone | None:
        logger.trace("Waiting for message in mail queue")
        email = await self.recvq.get()
        if not email:
            logger.trace("No message in mail queue")
            return

        reply_task = BotReplyTask(
            mailbot=self.mailbot,
            email=email.parsed(),
            send_queue=self.sendq,
            retries=self.retries,
        )
        reply_runner = reply_task.runner()
        reply_runner.start()
        if self.one_at_a_time:
            await reply_runner.wait()


def make_bot_reply_spawn_task(
    config: ReplyConfig,
    recv_queue: AsyncQueue[IMAPRawMessage],
    send_queue: AsyncQueue[SimpleEmailMessage],
):
    return BotReplySpawnTask(
        make_mailbot(
            config.models,
            configurable_fields=config.chat_model_configurable_fields,
        ),
        recv_queue=recv_queue,
        send_queue=send_queue,
    )



from __future__ import annotations

import datetime
import email
from dataclasses import dataclass
from email.message import EmailMessage
from typing import Self

from imap_tools.message import MailMessage
from imap_tools.utils import EmailAddress


@dataclass(slots=True)
class SimpleEmailMessage:
    """
    Used for composing simple plaintext messages.
    """

    addr_from: EmailAddress
    addrs_to: tuple[EmailAddress, ...]
    subject: str
    body: str
    date: datetime.datetime
    in_reply_to: str | None = None

    def __str__(self) -> str:
        from_str = self.addr_from.full
        to_str = ", ".join([a.full for a in self.addrs_to])
        return f"From: {from_str}\nTo: {to_str}\nSubject: {self.subject}\n{self.body}"

    def summary(self) -> str:
        ato = ", ".join([a.email for a in self.addrs_to])
        parts = ["Email", f"From: {self.addr_from.email}", f"To: {ato}"]
        if self.date:
            parts.append(f"Date: {self.date}")
        if self.in_reply_to:
            parts.append(f"In-Reply-To: {self.in_reply_to}")
        return " ".join(parts)

    def to_email_message(self) -> EmailMessage:
        msg = EmailMessage()
        msg["From"] = self.addr_from.full
        msg["To"] = ", ".join([a.full for a in self.addrs_to])
        msg["Subject"] = self.subject
        msg["Date"] = self.date.strftime("%a, %d %b %Y %H:%M:%S %z")
        msg["In-Reply-To"] = self.in_reply_to
        msg.set_content(self.body)
        return msg


@dataclass(slots=True)
class IMAPRawMessage:
    """
    Raw data from fetching emails with imap_tools.
    """

    message_data: bytes
    uid_data: bytes
    flag_data: list[bytes]
    uid: str | None = None

    @classmethod
    def from_fetch(cls, fetch_data: list) -> Self:
        message_data, uid_data, flag_data = MailMessage._get_message_data_parts(fetch_data)
        return cls(message_data=message_data, uid_data=uid_data, flag_data=flag_data)

    def parsed(self) -> IMAPMessage:
        return IMAPMessage(self)


class IMAPMessage(MailMessage):
    """
    Parsed email message from imap_tools.
    """

    def __init__(self, raw_email: IMAPRawMessage):
        self._raw_email = raw_email
        self.obj = email.message_from_bytes(raw_email.message_data)

    @property
    def _raw_message_data(self) -> bytes:
        return self._raw_email.message_data

    @property
    def _raw_flag_data(self) -> list[bytes]:
        return self._raw_email.flag_data

    @property
    def _raw_uid_data(self) -> bytes:
        return self._raw_email.uid_data

    @property
    def addr_from(self) -> EmailAddress:
        if not self.from_values:
            raise ValueError("message has no From address")
        return self.from_values

    @property
    def addrs_to(self) -> tuple[EmailAddress, ...]:
        return self.to_values

    @property
    def in_reply_to(self) -> str | None:
        return self.obj.get("In-Reply-To")

    @property
    def references(self) -> str | None:
        return self.obj.get("References")

    @property
    def message_id(self) -> str | None:
        return self.obj.get("Message-Id")

    def __str__(self) -> str:
        from_str = self.from_values.full if self.from_values else "unknown"
        to_str = ", ".join([a.full for a in self.addrs_to])
        return f"From: {from_str}\nTo: {to_str}\nSubject: {self.subject}\n{self.text}"

    def summary(self) -> str:
        ato = ", ".join([a.email for a in self.addrs_to])
        parts = ["Email", f"From: {self.addr_from.email}", f"To: {ato}"]
        if self.date:
            parts.append(f"Date: {self.date}")
        if self.message_id:
            parts.append(f"Message-Id: {self.message_id}")
        return " ".join(parts)

    def create_reply(self, addr_from: EmailAddress, body: str) -> SimpleEmailMessage:
        return SimpleEmailMessage(
            addr_from=addr_from,
            addrs_to=(self.addr_from,),
            subject=f"Re: {self.subject}",
            body=body,
            date=datetime.datetime.now(),
            in_reply_to=self.message_id,
        )



"""
Task and TaskRunner are used to define and run sync and async tasks.

Tasks are defined by subclassing SyncTask or AsyncTask and overriding the run method.
TaskRunners manage the execution of the tasks.

A TaskRunner executes the run method of its Task one or more times,
until one of the following occurs:
 - the task signals completion by returning TaskDone
 - the task is cancelled or stopped
 - an exception is raised and not caught by Task.handle_exception

Below are some examples for defining tasks with common execution patterns.

A one-shot task returns TaskDone immediately (but this should just be a function):
>>> class OneShotTask(AsyncTask[int]):
...     async def run(self) -> TaskDone[int]:
...         await asyncio.sleep(5)
...         return TaskDone(42)
>>>
>>> res = await OneShotTask().runner().start().result()

A long-lived "worker" task never returns TaskDone:
>>> class WorkerTask(AsyncTask[int]):
...     async def run(self) -> TaskDone[None]:
...         task = await self.task_queue.get()
...         result = self.do_some_work(task)
...         await self.result_queue.put(result)
>>>
>>> WorkerTask().runner().start()

A periodic task never returns TaskDone, and is started with an interval:
>>> class PeriodicTask(AsyncTask[int]):
...     async def run(self) -> TaskDone[None]:
...         print(f"The time is now: {time.time()}")
>>>
>>> PeriodicTask().runner().start(interval=1)

An iterative task returns TaskDone when its result is ready
>>> class IterativeTask(AsyncTask[int]):
...     async def run(self) -> TaskDone[int]:
...         self.count += 1
...         if self.count >= 5:
...             return TaskDone(self.count)
>>>
>>> task = IterativeTask().runner().start(interval=1)
>>> res = await task.result()

SyncTasks are run asynchronously using a ThreadPoolExecutor or ProcessPoolExecutor,
while AsyncTasks are run directly in asyncio.

Although you should probably just use asyncio.Task, this module adds a couple of things on top:
- TaskRunner tries to paper over the differences between sync and async units of work,
  though when running sync tasks you must still pay attention to thread/process-safety and the GIL
- As an alternative to cancel(), TaskRunner provides a stop() method, which acts as a soft-cancel.
  It lets the current Task.run() call finish before cancelling. For some tasks it makes it
  easier to reason about the task's lifecycle, compared to cancel(), which can
  raise CancelledError anywhere await is used. However stop() is not guaranteed
  to stop a task anytime soon, if it's waiting on I/O that may never come.

Beware that due to the Python GIL, running CPU-bound tasks in a ThreadPoolExecutor
still blocks the entire interpreter. Generally ThreadPoolExecutor is appropriate
for IO-bound tasks, while ProcessPoolExecutor is appropriate for CPU-bound tasks,
but has higher overhead.
"""

from __future__ import annotations

import abc
import asyncio
import time
from collections import defaultdict
from concurrent.futures import Executor
from typing import Generic, Self, TypeVar

from loguru import logger

_last_task_id: dict[type, int] = defaultdict(lambda: 0)


def get_next_task_id(cls: type) -> int:
    global _last_task_id
    _last_task_id[cls] += 1
    return _last_task_id[cls]


def default_task_name(obj: object) -> str:
    return f"{obj.__class__.__name__}.{get_next_task_id(obj.__class__)}"


T = TypeVar("T")


class TaskDone(Generic[T]):
    """
    Wrapper for the final result of a task, signalling that it should not be run again.
    """

    def __init__(self, result: T):
        self.result = result


class Task(abc.ABC, Generic[T]):
    def __init__(self, name: str | None = None):
        self._name = name or default_task_name(self)

    @abc.abstractmethod
    def runner(self, executor: Executor | None = None) -> TaskRunner[T]:
        pass

    def handle_exception(self, exc: Exception):
        """
        Called when an exception occurs during task execution.

        If the exception is re-raised, execution will be stopped.
        (The default implementation always re-raises.)

        Args:
            exc: The exception that was raised
        """
        logger.exception("Exception in task {}", self._name, exc_info=exc)
        raise exc

    def on_stopped(self):
        """
        Called after the task is stopped.

        By default, this method calls on_cancelled().
        """
        self.on_cancelled()

    def on_cancelled(self):
        """
        Called after the task is cancelled.

        It is also called when a task is stopped, unless on_stopped is overridden.

        By default, this method just logs that the task was cancelled.
        """
        logger.info("Task {} received asyncio.CancelledError", self._name)


class SyncTask(Task[T]):
    """
    SyncTask represents a task that can be run synchronously,
    or asynchronously in a thread or process pool.

    Subclasses must implement the run method to define the task's behavior.

    Optionally, subclasses can override handle_exception and on_cancelled
    to provide custom exception handling and cleanup operations.
    """

    @abc.abstractmethod
    def run(self) -> TaskDone[T] | None:
        """
        The main task function. Must be overriden by concrete Tasks.
        """
        pass

    def runner(self, executor: Executor | None = None) -> SyncTaskExecutorRunner[T]:
        """
        Get runner for this task.

        The same task can have multiple runners, but the task instance will
        be shared between runners.

        Creating multiple runners for the same task instance is only
        recommended for stateless tasks.
        """
        return SyncTaskExecutorRunner(self, executor)


class AsyncTask(Task[T]):
    """
    AsyncTask represents a task that can be run asynchronously.

    Subclasses must implement the run method to define the task's behavior.

    Optionally, subclasses can override handle_exception and on_cancelled
    to provide custom exception handling and cleanup operations.
    """

    @abc.abstractmethod
    async def run(self) -> TaskDone[T] | None:
        """
        The main task function. Must be overriden by concrete Tasks.
        """
        pass

    def runner(self, executor: Executor | None = None) -> AsyncTaskRunner[T]:
        """
        Get runner for this task.

        The same task can have multiple runners, but the task instance will
        be shared between runners.

        Creating multiple runners for the same task instance is only
        recommended for stateless tasks.

        The executor parameter has no effect for AsyncTask, the task runs
        with asyncio directly, it is only there to match the signature of
        Task/SyncTask.
        """
        if executor is not None:
            logger.warning("An executor was passed to AsyncTask; it will be ignored")
        return AsyncTaskRunner(self)


STOPPED_MESSAGE = "TaskRunner.STOPPED"


class StoppedError(asyncio.CancelledError):
    def __init__(self):
        super().__init__(STOPPED_MESSAGE)


class TaskRunner(abc.ABC, Generic[T]):
    """
    TaskRunner handles running a Task one or more times, until
    the Task returns a final result wrapped in TaskDone, or is interrupted.

    These are the ways a task can get interrupted:
    - immediately, when an exception is raised and not caught by Task.handle_exception
    - soon, when the task is cancelled
    - eventually, when the task is stopped

    See SyncTaskRunner and AsyncTaskRunner for concrete implementations
    which run SyncTask and AsyncTask respectively.
    """

    def __init__(self, task: Task[T]):
        self.run_until_done_task: asyncio.Task | None = None
        self.task = task
        self.name = f"{default_task_name(self)}<{task._name}>"

        self.done: TaskDone[T] | None = None
        self.stopped: bool = False
        self.cancelled: bool = False
        self.exception: Exception | None = None
        self.waiting_for_interval: bool = False

    @property
    def is_finished(self) -> bool:
        return bool(self.done or self.stopped or self.cancelled or self.exception)

    async def wait_for_interval(self, duration: float):
        self.waiting_for_interval = True
        await asyncio.sleep(duration)
        self.waiting_for_interval = False

    @abc.abstractmethod
    async def _run_task_async(self) -> TaskDone[T] | None:
        pass

    async def _run_once(self) -> None:
        # Should not happen, but just in case
        assert not self.is_finished, f"{self.name} is finished"
        try:
            task_output = await self._run_task_async()
            if isinstance(task_output, TaskDone):
                logger.success("Task {} is done", self.name)
                self.done = task_output

        # asyncio.CancelledError is a subclass of BaseException, not Exception
        # so it is not caught here. it is handled by _run_until_done
        except Exception as exc:
            try:
                self.task.handle_exception(exc)
            except Exception as inner_exc:
                self.exception = exc
                raise inner_exc from exc

    async def _run_until_done(self, interval: float | None = None) -> T:
        try:
            last_call_t = 0.0
            while True:
                if interval:
                    until_next_call_t = last_call_t + interval - time.time()
                    if until_next_call_t > 0.0:
                        logger.trace(
                            "Task {} waiting {:0.5f} seconds until next call",
                            self.name,
                            until_next_call_t,
                        )
                        await self.wait_for_interval(until_next_call_t)

                last_call_t = time.time()
                await self._run_once()

                # Some of these should not happen, but just in case
                if self.done is not None:
                    return self.done.result
                if self.stopped:
                    raise StoppedError()
                if self.cancelled:
                    raise asyncio.CancelledError()
                if self.exception:
                    raise self.exception

        # We don't catch Exception here, it is caught in _run_once
        # If an Exception reaches here it means it was not handled by Task.handle_exception
        # and should be propagated, or it was raised outside of Task.run
        # Task.handle_exception should not be called for an exception that was raised
        # by TaskRunner's own code rather than Task.run
        except asyncio.CancelledError as e:
            if e.args and e.args[0] == STOPPED_MESSAGE:
                self.stopped = True
                self.task.on_stopped()
            else:
                self.cancelled = True
                self.task.on_cancelled()
            raise

    def start(self, interval: float | None = None) -> Self:
        """
        Start the task, running it one or more times until done, cancelled,
        stopped, or interrupted by an unhandled exception.

        Args:
            interval: minimum delay between task executions, in seconds (default: None)
        """
        assert not self.run_until_done_task, f"{self.name} was already started"

        logger.info("Start running task {} with interval={}", self.name, interval)
        self.run_until_done_task = asyncio.create_task(
            self._run_until_done(interval),
            name=self.name,
        )
        return self

    def stop(self):
        """
        Stops future calls to Task.run.
        Waits for Task.run to return if it is currently executing.

        To interrupt Task.run, call TaskRunner.cancel() instead.
        """
        assert self.run_until_done_task and not self.is_finished, f"{self.name} is not ongoing"

        logger.info("Stopping task {}", self.name)
        self.stopped = True
        if self.waiting_for_interval:
            logger.info("Task {} is currently sleeping, cancelling it", self.name)
            self.run_until_done_task.cancel(STOPPED_MESSAGE)

    def cancel(self):
        """
        Cancel the task immediately, interrupting the current execution
        by raising asyncio.CancelledError.
        """
        assert self.run_until_done_task and not self.is_finished, f"{self.name} is not ongoing"

        logger.info("Cancelling task {}", self.name)
        self.cancelled = True
        self.run_until_done_task.cancel()

    async def shutdown(self, deadline: int | None = None):
        """
        Stop the task and wait for it to finish.

        Args:
            deadline: maximum time to wait for the task to finish, in seconds
                after which the task will be cancelled (default: None)
        """
        self.stop()
        if deadline is not None:
            try:
                await asyncio.wait_for(self.wait(), timeout=deadline)
            except asyncio.TimeoutError:
                self.cancel()
                await self.wait()
        else:
            await self.wait()

    async def result(self) -> T:
        """
        Wait and get result of the task, or raises exceptions, including
        asyncio.CancelledError.
        """
        assert self.run_until_done_task, f"{self.name} was not started"
        return await self.run_until_done_task

    async def wait(self) -> None:
        """
        Wait for the task to finish. Does not return result or raise.
        """
        assert self.run_until_done_task, f"{self.name} was not started"
        try:
            await self.run_until_done_task
        except asyncio.CancelledError:
            pass
        except Exception:
            pass


class AsyncTaskRunner(TaskRunner[T]):
    """
    AsyncTaskRunner runs an AsyncTask.

    See documentation for TaskRunner for more details on the TaskRunner interface.
    """

    def __init__(self, task: AsyncTask[T]):
        super().__init__(task)
        self.task = task

    async def _run_task_async(self) -> TaskDone[T] | None:
        return await self.task.run()


class SyncTaskExecutorRunner(TaskRunner[T]):
    """
    SyncTaskExecutorRunner runs a blocking SyncTask "asynchronously",
    using a concurrent.futures Executor.

    See documentation for TaskRunner for more details on the TaskRunner interface.

    If executor is not passed in, the event loop's default executor is used,
    which is normally a ThreadPoolExecutor.

    The loop's default executor can be changed by calling loop.set_default_executor.

    Depending on the type of executor used, the implementation of SyncTask.run
    must be thread-safe or process-safe.

    Due to the GIL, CPU-bound tasks block the event loop when run in a thread pool.
    """

    def __init__(self, task: SyncTask[T], executor: Executor | None = None):
        super().__init__(task)
        self.task = task
        self.executor = executor

    async def _run_task_async(self) -> TaskDone[T] | None:
        """
        Execute the blocking SyncTask.run method asynchronously by
        using the runner's executor.
        """
        loop = asyncio.get_running_loop()
        return await loop.run_in_executor(self.executor, self.task.run)



from __future__ import annotations

import datetime
import re
from pathlib import Path
from typing import Annotated, Any, List, Literal, TypeVar

import yaml
from annotated_types import Ge, Le
from config_path import ConfigPath
from pydantic import (
    ConfigDict,
    EmailStr,
    Field,
    NonNegativeInt,
    PositiveFloat,
    PositiveInt,
    SecretStr,
    model_validator,
)
from pydantic_settings import BaseSettings, SettingsConfigDict, YamlConfigSettingsSource
from pydantic_settings.sources import PydanticBaseSettingsSource

from llmailbot.duration import parse_duration
from llmailbot.enums import (
    EncryptionMode,
    FilterMode,
    VerifyMode,
)


class ConfigError(ValueError):
    pass


T = TypeVar("T")
type Opt[T] = T | None


Port = Annotated[int, Ge(1), Le(65535)]
Temperature = Annotated[float, Ge(0.0), Le(1.0)]


def snake_to_camel_case(snake_str: str) -> str:
    return "".join(word.title() for word in snake_str.split("_"))


def camel_to_snake_case(camel_str: str) -> str:
    # Insert underscore before uppercase letters and convert to lowercase
    snake_str = re.sub(r"(?<!^)(?=[A-Z])", "_", camel_str).lower()
    return snake_str


def yaml_config_locations() -> List[Path]:
    unix_common = Path.home() / ".config" / "llmailbot" / "config.yaml"
    os_convention = ConfigPath("llmailbot", "pigeonland.net", ".yaml").saveFilePath(mkdir=False)
    return [Path("./config.yaml"), unix_common, Path(os_convention)]


def secrets_dirs():
    paths = [Path("/run/secrets"), Path("/var/run/llmailbot/secrets")]
    return [p for p in paths if p.exists()]


class RootSettings(BaseSettings):
    model_config = SettingsConfigDict(
        alias_generator=snake_to_camel_case,
        case_sensitive=False,
        extra="ignore",
        secrets_dir=secrets_dirs(),
        yaml_file=yaml_config_locations(),
        populate_by_name=True,
    )

    @classmethod
    def settings_customise_sources(
        cls,
        settings_cls: type[BaseSettings],
        init_settings: PydanticBaseSettingsSource,
        env_settings: PydanticBaseSettingsSource,
        dotenv_settings: PydanticBaseSettingsSource,
        file_secret_settings: PydanticBaseSettingsSource,
    ) -> tuple[PydanticBaseSettingsSource, ...]:
        if cls.model_config.get("yaml_file"):
            yaml_settings = YamlConfigSettingsSource(settings_cls)
            return init_settings, yaml_settings, file_secret_settings
        else:
            return init_settings, file_secret_settings

    def dump_yaml(self) -> str:
        return yaml.dump(self.model_dump(mode="json", by_alias=True), sort_keys=False)


class SubSettings(BaseSettings):
    model_config = SettingsConfigDict(
        alias_generator=snake_to_camel_case,
        case_sensitive=False,
        extra="ignore",
        populate_by_name=True,
    )


DEFAULT_IMAP_ENCRYPTION = {
    143: EncryptionMode.STARTTLS,
    993: EncryptionMode.SSL_TLS,
}

DEFAULT_SMTP_ENCRYPTION = {
    25: EncryptionMode.NONE,
    587: EncryptionMode.STARTTLS,
    465: EncryptionMode.SSL_TLS,
}


class SMTPConfig(SubSettings):
    username: str = Field(...)
    password: SecretStr = Field(...)
    server: str = Field(...)
    port: Port = Field(465)
    encryption: Opt[EncryptionMode] = Field(None)

    @model_validator(mode="after")
    def validate_encryption(self) -> SMTPConfig:
        if self.encryption is None:
            try:
                self.encryption = DEFAULT_SMTP_ENCRYPTION[self.port]
            except KeyError as e:
                raise ConfigError(
                    f"Cannot infer encryption mode for non-standard SMTP port {self.port}. "
                    f"Please specify 'Encryption' explicitly."
                ) from e

        return self


class IMAPConfig(SubSettings):
    # Mail server settings
    username: str = Field(...)
    password: SecretStr = Field(...)
    server: str = Field(...)
    port: Port = Field(993)
    encryption: Annotated[Opt[EncryptionMode], Field()] = None

    watch_folder: Annotated[str, Field()] = "INBOX"
    replied_folder: Annotated[str | None, Field()] = "LLMailBot/Processed"
    blocked_folder: Annotated[str | None, Field()] = "LLMailBot/Blocked"
    max_fetch_rate: Annotated[PositiveFloat, Field()] = 1.0
    idle_timeout: Annotated[PositiveInt, Field()] = 30

    @model_validator(mode="after")
    def validate_encryption(self) -> IMAPConfig:
        if self.encryption is None:
            try:
                self.encryption = DEFAULT_IMAP_ENCRYPTION[self.port]
            except KeyError as e:
                raise ConfigError(
                    f"Cannot infer encryption mode for non-standard IMAP port {self.port}. "
                    f"Please specify 'Encryption' explicitly."
                ) from e

        return self


class MemoryQueueSettings(SubSettings):
    queue_type: Annotated[Literal["Memory"], Field(alias="Type")] = "Memory"
    max_size: Annotated[NonNegativeInt, Field()] = 0
    timeout: Annotated[PositiveInt, Field()] = 10


class RedisConfig(SubSettings):
    host: Annotated[str, Field()] = "localhost"
    port: Annotated[Port, Field()] = 6379
    db: Annotated[NonNegativeInt, Field()] = 0
    username: Annotated[Opt[str], Field()] = None
    password: Annotated[Opt[str], Field()] = None


class RedisQueueSettings(RedisConfig):
    queue_type: Annotated[Literal["Redis"], Field(alias="Type")] = "Redis"
    key: str = Field(...)
    timeout: Annotated[NonNegativeInt, Field()] = 10


type QueueSettings = MemoryQueueSettings | RedisQueueSettings


class RateLimitConfig(SubSettings):
    limit: PositiveInt = Field(...)
    window: Annotated[str, Field()] = "1 hour"

    _window_timedelta: datetime.timedelta

    @model_validator(mode="after")
    def validate_window(self) -> RateLimitConfig:
        self._window_timedelta = parse_duration(self.window)
        return self


class FilterHeaderConfig(SubSettings):
    header: str = Field(...)
    values: List[str] = Field(default_factory=list)
    mode: Annotated[FilterMode, Field()] = FilterMode.ALLOWLIST
    verify: Annotated[VerifyMode, Field()] = VerifyMode.ALWAYS


class SecurityConfig(SubSettings):
    rate_limit: RateLimitConfig = Field(
        default_factory=lambda: RateLimitConfig.model_validate({"Limit": 100, "Window": "1 day"})
    )
    rate_limit_per_sender: Annotated[Opt[RateLimitConfig], Field()] = None
    rate_limit_per_domain: Annotated[Opt[RateLimitConfig], Field()] = None

    # Secure default: dont allow any addresses
    allow_from: List[EmailStr] = Field(default_factory=list)
    allow_from_all_i_want_to_spend_it_all: Annotated[
        bool, Field(alias="AllowAllAddressesIReallyDontMindSpendingAllMyCredits")
    ] = False
    block_from: Annotated[Opt[List[EmailStr]], Field()] = None

    filter_headers: Annotated[Opt[List[FilterHeaderConfig]], Field()] = None

    verify_dkim: Annotated[VerifyMode, Field(alias="VerifyDKIM")] = VerifyMode.NEVER
    verify_mail_from: Annotated[VerifyMode, Field()] = VerifyMode.NEVER
    verify_x_mail_from: Annotated[VerifyMode, Field()] = VerifyMode.NEVER


class ChatModelConfig(SubSettings):
    model_config = ConfigDict(extra="allow")
    model: Opt[str] = None
    model_provider: Opt[str] = None
    max_tokens: PositiveInt = 1024
    temperature: Temperature = 0.2

    def chat_model_config(self) -> dict[str, Any]:
        config = self.model_dump()
        return {camel_to_snake_case(k): v for k, v in config.items()}


DEFAULT_SYSTEM_PROMPT = """
Your name is {name}. You are a helpful assistant.
A user is communicating with you using email.
Your email address is {bot_email}. The user's email address is {user_email}.
You must compose a reply in text/plain format.
Your reply must not include email headers such as To, From and Subject.
Messages in email conversations are often ordered from newest to oldest.
Reply to this email conversation:
""".lstrip()


class ModelSpec(SubSettings):
    name: str = Field(...)

    address: Annotated[Opt[EmailStr], Field()] = None
    address_regex: Annotated[Opt[str], Field()] = None

    max_input_length: Annotated[PositiveInt, Field()] = 5000
    system_prompt: str = Field(DEFAULT_SYSTEM_PROMPT)

    params: ChatModelConfig = Field(
        default_factory=lambda: ChatModelConfig(),
        alias="ChatModelConfig",
    )

    _address_regex: Opt[re.Pattern[str]]

    @model_validator(mode="after")
    def validate_exactly_one_email_addr(self) -> ModelSpec:
        if self.address_regex is not None and self.address is None:
            return self
        if self.address is not None and self.address_regex is None:
            return self
        raise ConfigError("exactly one of Address or AddressRegex must be set")

    @model_validator(mode="after")
    def validate_address_regex(self) -> ModelSpec:
        if self.address_regex is not None:
            try:
                self._address_regex = re.compile(self.address_regex)
            except re.error as e:
                raise ConfigError(f"invalid regex: {e}") from e
        else:
            self._address_regex = None
        return self

    def chat_model_config(self, email_addr: Opt[str] = None) -> dict[str, Any]:
        model_config = self.params.chat_model_config()
        if email_addr and self._address_regex:
            if m := self._address_regex.match(email_addr):
                for k, v in m.groupdict().items():
                    if v is not None:
                        model_config[k.lower()] = v

        return model_config


def default_queue() -> QueueSettings:
    return MemoryQueueSettings()


class FetchConfig(RootSettings):
    imap: IMAPConfig = Field(..., alias="IMAP")
    security: SecurityConfig = Field(default_factory=lambda: SecurityConfig())
    receive_queue: QueueSettings = Field(default_factory=default_queue)


class SendConfig(RootSettings):
    smtp: SMTPConfig = Field(..., alias="SMTP")
    send_queue: QueueSettings = Field(default_factory=default_queue)


class ReplyConfig(RootSettings):
    models: List[ModelSpec] = Field(...)
    chat_model_configurable_fields: Opt[set[str]] = Field(None)
    receive_queue: QueueSettings = Field(default_factory=default_queue)
    send_queue: QueueSettings = Field(default_factory=default_queue)

    @model_validator(mode="after")
    def normalize_chat_model_configurable_fields(self) -> ReplyConfig:
        if self.chat_model_configurable_fields is not None:
            self.chat_model_configurable_fields = set(
                [camel_to_snake_case(f) for f in self.chat_model_configurable_fields]
            )
        return self

    @model_validator(mode="after")
    def validate_unique_bot_addresses(self) -> ReplyConfig:
        addresses = [bot.address for bot in self.models]
        if len(addresses) != len(set(addresses)):
            raise ConfigError("Each mailbot must use a unique email address")
        return self



from enum import Enum

from loguru import logger

from llmailbot.email.model import IMAPMessage

try:
    import dkim
    import DNS
    from dkim import DKIMException
except ImportError as e:
    raise ImportError(
        "pydns and dkimpy are required for DKIM verification;"
        " install them with 'poetry install --extras dkim' or pip install llmailbot[dkim]"
    ) from e


class DnsError(Exception):
    pass


# The implementation in dkimpy doesn't follow CNAME records
# So it fails when the TXT record for DKIM is behind a CNAME
def get_dns_txt_recursive(name: str | bytes, timeout=5) -> bytes:
    """
    Return a TXT record associated with a DNS name.
    Follows CNAME records.
    """
    if isinstance(name, bytes):
        name = name.decode("utf-8")

    name = name.rstrip(".")
    response = DNS.DnsRequest(name, qtype="txt", timeout=timeout).req()

    if response is None:
        raise DnsError(f"DNS lookup for {name} returned no response")

    if not response.answers:
        raise DnsError(f"DNS lookup for {name} returned no answers")

    record_cname: str | None = None
    record_txt: bytes | None = None
    for answer in response.answers:
        rectype = answer["typename"].lower()
        if rectype == "txt":
            record_txt = b"".join(answer["data"])
        elif rectype == "cname":
            record_cname = answer["data"][0]

    if record_txt is not None:
        return record_txt

    if record_cname is not None:
        return get_dns_txt_recursive(record_cname, timeout)

    raise DnsError(f"DNS lookup for {name} returned no TXT or CNAME record")


class VerificationResult(Enum):
    PASS = 0
    FAIL = 1
    MISSING = 2


def verify_dkim_signatures(email: IMAPMessage, timeout: int = 5) -> VerificationResult:
    sigs = email.obj.get_all("DKIM-Signature")
    if not sigs:
        return VerificationResult.MISSING
    sig_indices = list(range(len(sigs)))

    verifier = dkim.DKIM(email._raw_message_data, logger=logger, timeout=timeout)
    for idx in sig_indices:
        if not verifier.verify(idx, dnsfunc=get_dns_txt_recursive):
            return VerificationResult.FAIL
    return VerificationResult.PASS


__all__ = ["verify_dkim_signatures", "DnsError", "VerificationResult", "DKIMException"]



import datetime
from enum import Enum


class LimitResult(Enum):
    OK = True
    EXCEEDED = False


# TODO: use bucket algorithm instead
# TODO: implement RedisRateLimiter
class RateLimiter:
    def __init__(self, duration: datetime.timedelta, limit: int):
        self.duration = duration
        self.limit = limit
        self._limit_count = 0
        self._limit_expiry = datetime.datetime.now() + self.duration

    def _reset(self, now: datetime.datetime) -> None:
        self._limit_count = 0
        self._limit_expiry = now + self.duration

    def _is_expired(self, now: datetime.datetime | None = None) -> bool:
        now = now or datetime.datetime.now()
        return now > self._limit_expiry

    def count(self) -> LimitResult:
        now = datetime.datetime.now()
        if self._is_expired(now):
            self._reset(now)

        self._limit_count += 1
        if self._limit_count > self.limit:
            return LimitResult.EXCEEDED
        return LimitResult.OK



import abc
import asyncio
from typing import Generic, TypeVar

T = TypeVar("T")


class SyncQueue(abc.ABC, Generic[T]):
    @abc.abstractmethod
    def put(self, message: T) -> None:
        pass

    @abc.abstractmethod
    def get(self) -> T | None:
        pass


class AsyncQueue(abc.ABC, Generic[T]):
    @abc.abstractmethod
    async def put(self, message: T) -> None:
        pass

    @abc.abstractmethod
    async def get(self) -> T | None:
        pass


class AsyncAdapter(AsyncQueue[T]):
    def __init__(self, queue: SyncQueue[T]):
        self.queue = queue

    async def put(self, message: T) -> None:
        await asyncio.to_thread(self.queue.put, message)

    async def get(self) -> T | None:
        return await asyncio.to_thread(self.queue.get)


def to_async_queue(q: SyncQueue[T] | AsyncQueue[T]) -> AsyncQueue[T]:
    if isinstance(q, AsyncQueue):
        return q
    return AsyncAdapter(q)



import sys
from enum import StrEnum

from loguru import logger

# Between WARNING and ERROR
logger.level("SECURITY", no=35, color="<red>", icon="🔒")


class LogLevel(StrEnum):
    TRACE = "TRACE"
    DEBUG = "DEBUG"
    INFO = "INFO"
    SUCCESS = "SUCCESS"
    WARNING = "WARNING"
    SECURITY = "SECURITY"
    ERROR = "ERROR"
    CRITICAL = "CRITICAL"


def setup_logging(log_file=None, log_level: str | LogLevel = LogLevel.INFO):
    log_level = LogLevel(log_level.upper())
    logger.remove()
    logger.add(log_file or sys.stderr, level=log_level.value)



import re
from enum import StrEnum
from functools import lru_cache


class CaseInsensitiveStrEnum(StrEnum):
    """
    A string enum that is whitespace-insensitive and case-insensitive when comparing values.
    """

    @staticmethod
    def _normalize(value: str) -> str:
        return re.sub(r"\s+", "", value).lower()

    @classmethod
    @lru_cache(maxsize=128)
    def _missing_(cls, value):
        if not isinstance(value, str):
            return None

        normalized_value = cls._normalize(value)
        for member in cls:
            if cls._normalize(member.value) == normalized_value:
                return member
        return None


class EncryptionMode(CaseInsensitiveStrEnum):
    """Encryption modes for SMTP or IMAP connections."""

    NONE = "none"
    STARTTLS = "starttls"
    SSL_TLS = "ssl/tls"


class WorkerType(CaseInsensitiveStrEnum):
    """Worker type for executor pools."""

    THREAD = "thread"
    PROCESS = "process"


class FilterMode(CaseInsensitiveStrEnum):
    """Specify if list is an allowlist or denylist."""

    ALLOWLIST = "allow list"
    DENYLIST = "deny list"


class VerifyMode(CaseInsensitiveStrEnum):
    NEVER = "never"
    IF_PRESENT = "if present"
    ALWAYS = "always"



# pyright: reportCallIssue=false
import functools
import sys

import aiorun
import click

from llmailbot.config import FetchConfig, ReplyConfig, SendConfig
from llmailbot.core import AppComponent, run_app
from llmailbot.logging import LogLevel, setup_logging


def handle_cli_exceptions(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as e:
            click.echo(f"Error: {str(e)}", err=True)
            sys.exit(1)

    return wrapper


@click.group()
@click.option("--config", "config_file", default=None, help="Configuration file (default: None)")
@click.option("--log-level", type=str, default=LogLevel.INFO, help="Log level (default: INFO)")
@click.option("--log-file", default=None, help="Log file (default: stderr)")
@handle_cli_exceptions
def cli(config_file: str | None, log_level: str, log_file: str | None):
    setup_logging(log_file, log_level)
    if config_file:
        FetchConfig.model_config["yaml_file"] = config_file
        ReplyConfig.model_config["yaml_file"] = config_file
        SendConfig.model_config["yaml_file"] = config_file


def indent(txt, indent_str="  ") -> str:
    return "\n".join(f"{indent_str}{line}" for line in txt.split("\n"))


@cli.group()
def config():
    pass


@config.command()
@click.argument("components", type=AppComponent, nargs=-1)
@handle_cli_exceptions
def show(components: list[AppComponent]):
    """
    Print loaded configuration in YAML format.
    """
    if not components:
        components = list(AppComponent)

    if AppComponent.FETCH in components:
        fetch_conf = FetchConfig()
        click.echo("Fetch configuration:")
        click.echo(indent(fetch_conf.dump_yaml()))

    if AppComponent.REPLY in components:
        reply_conf = ReplyConfig()
        click.echo("Reply configuration:")
        click.echo(indent(reply_conf.dump_yaml()))

    if AppComponent.SEND in components:
        send_conf = SendConfig()
        click.echo("Send configuration:")
        click.echo(indent(send_conf.dump_yaml()))


@config.command()
@handle_cli_exceptions
def example():
    """
    Write example configuration to ./config.yaml.
    """
    import importlib.resources
    from pathlib import Path

    # Check if config.yaml already exists
    config_path = Path("./config.yaml")
    if config_path.exists():
        click.echo(f"Error: {config_path.absolute()} already exists. Will not overwrite.", err=True)
        return

    # Get the example config from package resources
    example_config = (
        importlib.resources.files("llmailbot").joinpath("../config.example.yaml").read_text()
    )

    # Write to config.yaml
    with open(config_path, "w") as f:
        f.write(example_config)

    click.echo(f"Example configuration written to {config_path.absolute()}")


@config.command()
@handle_cli_exceptions
def interactive():
    """
    Interactively generate a config based on a series of prompts
    and write it to ./config.yaml.
    """
    from pathlib import Path

    import yaml

    # Check if config.yaml already exists
    config_path = Path("./config.yaml")
    if config_path.exists():
        click.echo(f"Error: {config_path.absolute()} already exists. Will not overwrite.", err=True)
        return

    # Initialize the config dictionary
    config = {}

    # IMAP Configuration
    click.echo("\nIMAP Configuration:")
    click.echo("=================\n")
    click.echo("⚠️  WARNING: Use a dedicated email account for LLMailBot. ⚠️")
    click.echo("⚠️  The bot may delete emails from this account. ⚠️\n")

    imap_config = {}
    imap_config["Server"] = click.prompt("IMAP Server", type=str)
    imap_config["Port"] = click.prompt("IMAP Port", type=int, default=993)
    imap_config["Username"] = click.prompt("IMAP Username", type=str)
    imap_password = click.prompt("IMAP Password", hide_input=True, confirmation_prompt=True)
    imap_config["Password"] = imap_password

    # Email fetch configuration
    imap_config["WatchFolder"] = click.prompt(
        "Folder to watch for incoming emails", default="INBOX"
    )
    imap_config["RepliedFolder"] = click.prompt(
        "Move replies to folder", default="LLMailBot/Processed"
    )
    imap_config["BlockedFolder"] = click.prompt(
        "Folder for blocked emails", default="LLMailBot/Blocked"
    )

    # SMTP Configuration
    click.echo("\nSMTP Configuration:")
    click.echo("=================\n")
    click.echo("This is used to send replies to emails.\n")

    smtp_config = {}
    # Try to suggest an SMTP server based on the IMAP server
    suggested_smtp = None
    if imap_config["Server"].startswith("imap."):
        # If IMAP server is imap.example.com, suggest smtp.example.com
        domain = imap_config["Server"][5:]  # Remove "imap."
        suggested_smtp = f"smtp.{domain}"
    smtp_config["Server"] = click.prompt("SMTP Server", type=str, default=suggested_smtp)
    smtp_config["Port"] = click.prompt("SMTP Port", type=int, default=587)
    smtp_config["Username"] = click.prompt(
        "SMTP Username", type=str, default=imap_config["Username"]
    )
    use_same_password = click.confirm("Use same password as IMAP?", default=True)
    if use_same_password:
        smtp_config["Password"] = imap_password
    else:
        smtp_config["Password"] = click.prompt(
            "SMTP Password", hide_input=True, confirmation_prompt=True
        )

    # Model Configuration
    click.echo("\nLLM Configuration:")
    click.echo("================\n")
    click.echo("This configures which LLM to use for generating replies.\n")

    # Create Models list
    models = []
    model_spec = {}

    # Get bot email address
    bot_email = click.prompt("Bot's email address", type=str, default=imap_config["Username"])
    model_spec["Address"] = bot_email

    # Set a name for the model
    model_spec["Name"] = click.prompt("Bot name", default="LLMailBot")

    # Set max input length
    model_spec["MaxInputLength"] = click.prompt("Maximum input length", default=10000, type=int)

    # Configure the chat model
    chat_model_config = {}
    model_provider = click.prompt(
        "Model Provider",
        type=click.Choice(["openai", "anthropic", "google", "ollama", "other"]),
        default="openai",
    )

    if model_provider == "openai":
        chat_model_config["ModelProvider"] = "openai"
        chat_model_config["Model"] = click.prompt("Model Name", default="gpt-4o")
        click.echo("\nPlease set your API key as an environment variable:")
        click.echo("export OPENAI_API_KEY=your_api_key_here")
        click.echo("\nThe API key will be loaded from the environment variable.")
    elif model_provider == "anthropic":
        chat_model_config["ModelProvider"] = "anthropic"
        chat_model_config["Model"] = click.prompt("Model Name", default="claude-3-5-sonnet-latest")
        click.echo("\nPlease set your API key as an environment variable:")
        click.echo("export ANTHROPIC_API_KEY=your_api_key_here")
        click.echo("\nThe API key will be loaded from the environment variable.")
    elif model_provider == "google":
        chat_model_config["ModelProvider"] = "google"
        chat_model_config["Model"] = click.prompt("Model Name", default="gemini-1.5-pro")
        click.echo("\nPlease set your API key as an environment variable:")
        click.echo("export GOOGLE_API_KEY=your_api_key_here")
        click.echo("\nThe API key will be loaded from the environment variable.")
    elif model_provider == "ollama":
        chat_model_config["ModelProvider"] = "ollama"
        chat_model_config["Model"] = click.prompt("Model Name", default="llama3")
        chat_model_config["BaseUrl"] = click.prompt(
            "Ollama Base URL", default="http://localhost:11434"
        )
    else:  # other
        provider_name = click.prompt("Provider Name", type=str)
        chat_model_config["ModelProvider"] = provider_name
        chat_model_config["Model"] = click.prompt("Model Name", type=str)
        if click.confirm("Does this provider require a custom base URL?", default=False):
            chat_model_config["BaseUrl"] = click.prompt("Base URL", type=str)

    # Set other model parameters
    chat_model_config["MaxTokens"] = click.prompt("Maximum tokens", default=2048, type=int)
    chat_model_config["Temperature"] = click.prompt("Temperature", default=0.2, type=float)

    # Add chat model config to model spec
    model_spec["ChatModelConfig"] = chat_model_config

    # Add model spec to models list
    models.append(model_spec)

    # Security Configuration
    click.echo("\nSecurity Configuration:")
    click.echo("=====================\n")
    click.echo("This configures which email addresses are allowed to use the bot.\n")

    security_config = {}
    security_config["AllowFrom"] = []
    user_email = click.prompt("Email address allowed to use the bot", type=str)
    security_config["AllowFrom"].append(user_email)

    while click.confirm("Add another allowed email address?", default=False):
        email = click.prompt("Email address", type=str)
        security_config["AllowFrom"].append(email)

    # Assemble the final config
    config["IMAP"] = imap_config
    config["SMTP"] = smtp_config
    config["Models"] = models
    config["Security"] = security_config

    # Add ChatModelConfigurableFields
    config["ChatModelConfigurableFields"] = [
        "Model",
        "ModelProvider",
        "MaxTokens",
        "Temperature",
        "BaseUrl",
    ]

    # Write to config.yaml
    with open(config_path, "w") as f:
        yaml.dump(config, f, default_flow_style=False, sort_keys=False)

    click.echo(f"\nConfiguration written to {config_path.absolute()}")
    click.echo("You can now run LLMailBot with: llmailbot run")


@cli.command()
@click.argument("components", type=AppComponent, nargs=-1)
@click.option("--threads", type=int, default=2, help="Number of threads to use (default: 2)")
@handle_cli_exceptions
def run(components: list[AppComponent], threads: int):
    """
    Start the mail bot.

    By default, runs all components (fetch, reply, send).

    Specify components to run only those.
    """

    aiorun.run(
        run_app(components=components),
        stop_on_unhandled_errors=True,
        timeout_task_shutdown=10,
        executor_workers=threads,
    )


if __name__ == "__main__":
    cli()



import abc
import asyncio
from smtplib import SMTP, SMTP_SSL
from ssl import SSLContext
from typing import Iterable

from loguru import logger

from llmailbot.config import EncryptionMode, SMTPConfig
from llmailbot.email.model import SimpleEmailMessage
from llmailbot.queue.core import AsyncQueue, SyncQueue, to_async_queue
from llmailbot.taskrun import AsyncTask, TaskDone


def connect_smtp(
    config: SMTPConfig,
    ssl_context: SSLContext | None = None,
    timeout: int = 30,
) -> SMTP | SMTP_SSL:
    if config.encryption == EncryptionMode.STARTTLS:
        client = SMTP(host=config.server, port=config.port, timeout=timeout)
        client.starttls(context=ssl_context)
    elif config.encryption == EncryptionMode.SSL_TLS:
        client = SMTP_SSL(
            host=config.server, port=config.port, timeout=timeout, context=ssl_context
        )
    else:
        client = SMTP(host=config.server, port=config.port, timeout=timeout)

    client.login(config.username, config.password.get_secret_value())
    return client


class MailSender(abc.ABC):
    def __init__(self, smtp_config: SMTPConfig):
        self.smtp_config = smtp_config

    @abc.abstractmethod
    def send(self, emails: Iterable[SimpleEmailMessage]) -> None:
        pass


class SMTPSender(MailSender):
    def send(self, emails: Iterable[SimpleEmailMessage]) -> None:
        with connect_smtp(self.smtp_config) as client:
            for email in emails:
                msg = email.to_email_message()
                client.send_message(
                    msg=msg,
                    from_addr=email.addr_from.email,
                    to_addrs=[a.email for a in email.addrs_to],
                )
                client.quit()


class StdoutFakeMailSender(MailSender):
    def send(self, email: SimpleEmailMessage) -> None:
        print(str(email))


class SendMailTask(AsyncTask[None]):
    def __init__(self, sender: MailSender, queue: AsyncQueue[SimpleEmailMessage]):
        super().__init__(
            name=f"SendMail<{sender.smtp_config.username}@{sender.smtp_config.server}>"
        )
        self.name = self._name
        self.sender = sender
        self.mailq = queue

    async def run(self) -> TaskDone | None:
        # TODO: implement batching to avoid re-connecting for every email
        email = await self.mailq.get()
        if email:
            await asyncio.get_running_loop().run_in_executor(
                None, lambda: self.sender.send([email])
            )
            logger.success("Sent {}", email.summary())


def make_mail_sender(config: SMTPConfig) -> MailSender:
    return SMTPSender(config)


def make_mail_send_task(
    config: SMTPConfig,
    queue: AsyncQueue[SimpleEmailMessage] | SyncQueue[SimpleEmailMessage],
):
    sender = make_mail_sender(config)
    return SendMailTask(sender, to_async_queue(queue))



import asyncio
import pickle
from typing import Awaitable, Callable, cast

from loguru import logger

from llmailbot.config import RedisConfig

from .core import AsyncQueue, SyncQueue, T

try:
    import redis
    import redis.asyncio as aioredis
except ImportError as e:
    raise ImportError(
        "redis is required for Redis queue; install it with "
        "'poetry install --extras redis' or pip install llmailbot[redis]"
    ) from e


type Serialize[T] = Callable[[T], bytes]
type Deserialize[T] = Callable[[bytes], T]


class AsyncRedisQueue(AsyncQueue[T]):
    def __init__(
        self,
        conf: RedisConfig,
        key: str,
        timeout: int = 5,
        serialize: Serialize[T] = pickle.dumps,
        deserialize: Deserialize[T] = pickle.loads,
    ):
        self.redis = aioredis.Redis(
            host=conf.host,
            port=conf.port,
            db=conf.db,
            username=conf.username,
            password=conf.password,
        )
        self.key = key
        self.serialize = serialize
        self.deserialize = deserialize
        self.timeout = timeout

    async def put(self, message: T) -> None:
        await asyncio.wait_for(
            cast(Awaitable[int], self.redis.lpush(self.key, self.serialize(message))),
            timeout=self.timeout,
        )

    async def get(self) -> T | None:
        vals = await cast(
            Awaitable[list[bytes]], self.redis.brpop([self.key], timeout=self.timeout)
        )
        if not vals:
            return None
        logger.trace("Got {} values from Redis queue", repr([type(val) for val in vals]))
        return self.deserialize(vals[1])


class SyncRedisQueue(SyncQueue[T]):
    def __init__(
        self,
        conf: RedisConfig,
        key: str,
        serialize: Serialize[T] = pickle.dumps,
        deserialize: Deserialize[T] = pickle.loads,
        timeout: int = 5,
    ):
        self.redis = redis.Redis(
            host=conf.host,
            port=conf.port,
            db=conf.db,
            username=conf.username,
            password=conf.password,
        )
        self.key = key
        self.serialize = serialize
        self.deserialize = deserialize
        self.timeout = timeout

    def put(self, message: T) -> None:
        self.redis.lpush(self.key, self.serialize(message))

    def get(self) -> T | None:
        vals = cast(list[bytes], self.redis.brpop([self.key], timeout=self.timeout))
        if not vals:
            return None
        logger.trace("Got {} values from Redis queue", repr([type(val) for val in vals]))
        return self.deserialize(vals[1])



from llmailbot.config import (
    MemoryQueueSettings,
    QueueSettings,
    RedisQueueSettings,
)
from llmailbot.queue.core import AsyncQueue, SyncQueue
from llmailbot.queue.memory import AsyncioQueue, ManagedMemoryQueue, MemoryQueue

try:
    from .redis import AsyncRedisQueue
except ImportError:
    AsyncRedisQueue = None

__all__ = [
    "SyncQueue",
    "make_queue",
    "MemoryQueue",
    "ManagedMemoryQueue",
    "AsyncRedisQueue",
]


def make_queue(config: QueueSettings) -> AsyncQueue:
    if isinstance(config, MemoryQueueSettings):
        return AsyncioQueue(maxsize=config.max_size, timeout=config.timeout)
    if isinstance(config, RedisQueueSettings):
        if AsyncRedisQueue is None:
            raise RuntimeError("RedisQueue not available - redis extra must be installed")
        return AsyncRedisQueue(
            conf=config,
            key=config.key,
            timeout=config.timeout,
        )



import asyncio
from enum import StrEnum
from typing import Awaitable, Iterable

from loguru import logger

from llmailbot.config import ConfigError, FetchConfig, QueueSettings, ReplyConfig, SendConfig
from llmailbot.email.fetch import make_mail_fetch_task
from llmailbot.email.model import IMAPRawMessage, SimpleEmailMessage
from llmailbot.email.send import make_mail_send_task
from llmailbot.mailbot import make_bot_reply_spawn_task
from llmailbot.queue import make_queue
from llmailbot.queue.core import AsyncQueue


class AppComponent(StrEnum):
    FETCH = "fetch"
    REPLY = "reply"
    SEND = "send"


_mail_recv_q: AsyncQueue[IMAPRawMessage] | None = None


def get_mail_recv_q(conf: QueueSettings | None) -> AsyncQueue[IMAPRawMessage]:
    global _mail_recv_q
    if conf is None:
        raise ConfigError("Receive queue not configured")
    if _mail_recv_q is None:
        _mail_recv_q = make_queue(conf)

    return _mail_recv_q


_mail_send_q: AsyncQueue[SimpleEmailMessage] | None = None


def get_mail_send_q(conf: QueueSettings | None) -> AsyncQueue[SimpleEmailMessage]:
    global _mail_send_q
    if conf is None:
        raise ConfigError("Send queue not configured")
    if _mail_send_q is None:
        _mail_send_q = make_queue(conf)

    return _mail_send_q


async def run_app(components: Iterable[AppComponent] | None = None):
    if not components:
        components = list(AppComponent)
    components = set(components)

    tasks: list[Awaitable[None]] = []

    if AppComponent.FETCH in components:
        fetch_conf = FetchConfig()  # pyright: ignore[reportCallIssue]
        tasks.append(
            make_mail_fetch_task(
                fetch_conf.imap,
                fetch_conf.security,
                get_mail_recv_q(fetch_conf.receive_queue),
            )
            .runner()
            .start(interval=1.0 / fetch_conf.imap.max_fetch_rate)
            .result()
        )

    if AppComponent.REPLY in components:
        reply_conf = ReplyConfig()  # pyright: ignore[reportCallIssue]
        if not reply_conf.models:
            raise ConfigError("No LLM model configured")

        tasks.append(
            make_bot_reply_spawn_task(
                reply_conf,
                get_mail_recv_q(reply_conf.receive_queue),
                get_mail_send_q(reply_conf.send_queue),
            )
            .runner()
            .start()
            .result()
        )

    if AppComponent.SEND in components:
        send_conf = SendConfig()  # pyright: ignore[reportCallIssue]
        tasks.append(
            make_mail_send_task(
                send_conf.smtp,
                get_mail_send_q(send_conf.send_queue),
            )
            .runner()
            .start()
            .result(),
        )

    logger.success("All tasks started")
    await asyncio.gather(*tasks)



import datetime
import re

RE_DURATION = re.compile(r"^(\d+)\s?(seconds?|sec|s|m|min|minutes?|h|hours?|d|days?)$")

ALT_UNITS = {
    "d": "days",
    "day": "days",
    "h": "hours",
    "hour": "hours",
    "m": "minutes",
    "min": "minutes",
    "minute": "minutes",
    "s": "seconds",
    "sec": "seconds",
    "second": "seconds",
}


def parse_duration(duration: str) -> datetime.timedelta:
    match = re.match(RE_DURATION, duration)
    if not match:
        raise ValueError(f"Invalid duration: {duration}")

    n, unit = match.groups()
    unit = ALT_UNITS.get(unit, unit)
    n = int(n)

    kwargs = {unit: n}
    return datetime.timedelta(**kwargs)



import asyncio
import multiprocessing
import multiprocessing.managers
import queue

from .core import AsyncQueue, SyncQueue, T


class MemoryQueue(SyncQueue[T]):
    def __init__(self, maxsize: int = 0, timeout: float = 5.0):
        self.msgq = queue.Queue(maxsize=maxsize)
        self.timeout = timeout

    def put(self, message: T) -> None:
        self.msgq.put(message, block=True, timeout=self.timeout)

    def get(self) -> T | None:
        try:
            return self.msgq.get(block=True, timeout=self.timeout)
        except queue.Empty:
            return None


_manager = None


def get_manager() -> multiprocessing.managers.SyncManager:
    global _manager
    if _manager is None:
        _manager = multiprocessing.Manager()
    return _manager


class ManagedMemoryQueue(SyncQueue[T]):
    def __init__(self, maxsize: int = 0, timeout: float = 5.0):
        self.msgq = get_manager().Queue(maxsize=maxsize)
        self.timeout = timeout

    def put(self, message: T) -> None:
        self.msgq.put(message, block=True, timeout=self.timeout)

    def get(self) -> T | None:
        try:
            return self.msgq.get(block=True, timeout=self.timeout)
        except queue.Empty:
            return None


class AsyncioQueue(AsyncQueue[T]):
    def __init__(self, maxsize: int = 0, timeout: float = 5.0):
        self.msgq = asyncio.Queue(maxsize=maxsize)
        self.timeout = timeout

    async def put(self, message: T) -> None:
        await asyncio.wait_for(self.msgq.put(message), timeout=self.timeout)

    async def get(self) -> T | None:
        try:
            return await asyncio.wait_for(self.msgq.get(), timeout=self.timeout)
        except TimeoutError:
            return None



import asyncio
from ssl import SSLContext
from typing import Any, Awaitable, Callable, cast, override

from imap_tools.mailbox import BaseMailBox, MailBox, MailBoxTls, MailBoxUnencrypted
from imap_tools.query import AND
from loguru import logger

from llmailbot.config import EncryptionMode, IMAPConfig, SecurityConfig
from llmailbot.email.model import IMAPRawMessage
from llmailbot.queue.core import (
    AsyncQueue,
    SyncQueue,
    to_async_queue,
)
from llmailbot.security import Action, SecurityFilter, make_security_filter
from llmailbot.taskrun import AsyncTask, SyncTask

MAILBOX_CLS = {
    EncryptionMode.NONE: MailBoxUnencrypted,
    EncryptionMode.STARTTLS: MailBoxTls,
    EncryptionMode.SSL_TLS: MailBox,
}


def connect_mailbox(
    config: IMAPConfig,
    ssl_context: SSLContext | None = None,
    timeout: int | None = None,
) -> BaseMailBox:
    kwargs = {
        "host": config.server,
        "port": config.port,
        "timeout": timeout,
    }
    if ssl_context is not None and config.encryption != EncryptionMode.NONE:
        kwargs["ssl_context"] = ssl_context

    if config.encryption is None:
        raise ValueError("Encryption mode cannot be None")
    cls: type[BaseMailBox] = MAILBOX_CLS[config.encryption]
    mailbox: BaseMailBox = cls(**kwargs)
    mailbox.login(
        config.username, config.password.get_secret_value(), initial_folder=config.watch_folder
    )
    mailbox.email_message_class = IMAPRawMessage.from_fetch  # pyright: ignore[reportAttributeAccessIssue]
    return mailbox


class MailboxWatcher(AsyncTask):
    """
    Long-lived watcher for an IMAP mailbox folder.
    """

    def __init__(
        self,
        config: IMAPConfig,
        callback: Callable[[BaseMailBox, IMAPRawMessage], Awaitable[Any]],
        ssl_context: SSLContext | None = None,
        timeout: int | None = None,
    ):
        super().__init__()
        self.config = config
        self.ssl_context = ssl_context
        self.timeout = timeout
        self.folder_delim = None
        self.move_to_folder = None
        self.blocked_folder = None
        self.mb = connect_mailbox(self.config, self.ssl_context, self.timeout)
        self.setup_folders()
        self.callback = callback
        self.uids = []

    async def fetch_next_uid(self) -> IMAPRawMessage | None:
        next_uid = self.uids.pop()
        logger.trace("Fetching message with UID {}", next_uid)
        messages_iter = await asyncio.to_thread(self.mb.fetch, AND(uid=next_uid))
        messages = cast(list[IMAPRawMessage], list(messages_iter))
        if not messages:
            logger.warning("Failed to fetch message with UID {}", next_uid)
            return
        if len(messages) > 1:
            logger.warning("Multiple messages found with UID {}", next_uid)
        message = messages[0]
        message.uid = next_uid
        return message

    @override
    async def run(self):
        # Keep fetching the UIDs we already have before polling for new ones
        if self.uids:
            message = await self.fetch_next_uid()
            if message is not None:
                logger.trace("Callback processing message with UID {}", message.uid)
                await self.callback(self.mb, message)
            return

        new_messages = self.mb.folder.status().get("MESSAGES", 0)
        logger.trace("New messages count: {}", new_messages)
        if new_messages > 0:
            new_uids = await asyncio.to_thread(self.mb.uids)
            self.uids.extend(new_uids)
            logger.trace("Fetched {} new uids", len(self.uids))
        else:
            logger.trace("No new messages, IDLE for {} seconds", self.config.idle_timeout)
            # IDLE timeout doesn't play well with asyncio
            # TODO: find solution to use IDLE without hanging the process or
            # messing up the IMAP client states
            await asyncio.sleep(self.config.idle_timeout)
            # await asyncio.to_thread(self.mb.idle.wait, self.config.idle_timeout)
            return

    @override
    def on_cancelled(self):
        try:
            # Safely close the mailbox connection
            # Don't try to stop IDLE explicitly as it may cause state issues
            logger.debug("Logging out from mailbox")
            self.mb.logout()
        except Exception as e:
            logger.error("Error during mailbox cleanup: {}", str(e))

    @override
    def handle_exception(self, exc: Exception):
        try:
            # Safely close the mailbox connection
            # Don't try to stop IDLE explicitly as it may cause state issues
            logger.exception("Logging out from mailbox due to exception", exc_info=exc)
            self.mb.logout()
        except Exception as e:
            logger.exception("Error during mailbox cleanup after exception", exc_info=e)
        raise exc

    def setup_folders(self):
        if self.folder_delim is None:
            self.folder_delim = self.mb.folder.list(self.config.watch_folder)[0].delim
            if self.config.replied_folder:
                self.move_to_folder = self.config.replied_folder.replace("/", self.folder_delim)
            if self.config.blocked_folder:
                self.blocked_folder = self.config.blocked_folder.replace("/", self.folder_delim)

        if self.config.replied_folder:
            if not self.mb.folder.list(self.config.replied_folder):
                self.mb.folder.create(self.config.replied_folder)
        if self.config.blocked_folder:
            if not self.mb.folder.list(self.config.blocked_folder):
                self.mb.folder.create(self.config.blocked_folder)


def move_or_delete(mb: BaseMailBox, uid: str | list[str], folder: str | None):
    if folder:
        mb.move(uid, folder)
    else:
        mb.delete(uid)


def filter_and_enqueue(
    q: AsyncQueue[IMAPRawMessage],
    secf: SecurityFilter | None,
    replied_folder: str | None,
    blocked_folder: str | None,
) -> Callable[[BaseMailBox, IMAPRawMessage], Awaitable[None]]:
    if secf is None:

        async def qcallback(mb: BaseMailBox, message: IMAPRawMessage):
            await q.put(message)
            move_or_delete(mb, cast(str, message.uid), replied_folder)
    else:

        async def qcallback(mb: BaseMailBox, message: IMAPRawMessage):
            res = secf.apply(message.parsed())
            if res == Action.ALLOW:
                await q.put(message)
                move_or_delete(mb, cast(str, message.uid), replied_folder)
            else:
                move_or_delete(mb, cast(str, message.uid), blocked_folder)

    return qcallback


def make_mail_fetch_task(
    config: IMAPConfig,
    sec_config: SecurityConfig,
    queue: SyncQueue[IMAPRawMessage] | AsyncQueue[IMAPRawMessage],
) -> AsyncTask[None] | SyncTask[None]:
    return MailboxWatcher(
        config,
        callback=filter_and_enqueue(
            to_async_queue(queue),
            make_security_filter(sec_config),
            config.replied_folder,
            config.blocked_folder,
        ),
    )



import abc
import datetime
import re
from enum import Enum
from typing import Iterable, NamedTuple

from loguru import logger

from llmailbot.config import FilterMode, SecurityConfig
from llmailbot.email.model import IMAPMessage
from llmailbot.enums import VerifyMode
from llmailbot.ratelimit import LimitResult, RateLimiter


class Action(Enum):
    BLOCK = 0
    ALLOW = 1


class RuleResult(NamedTuple):
    action: Action
    reason: str | None

    @property
    def is_blocked(self) -> bool:
        return self.action == Action.BLOCK


class Rule(abc.ABC):
    @abc.abstractmethod
    def check(self, email: IMAPMessage) -> RuleResult:
        pass


VerifyDKIM: type[Rule] = None  # pyright: ignore[reportAssignmentType]

try:
    from llmailbot.dkim import DKIMException, DnsError, VerificationResult, verify_dkim_signatures

    class VerifyDKIM(Rule):
        """
        Blocks emails with an invalid DKIM signature.

        In strict mode, blocks emails without a DKIM signature.

        DKIM verification is most likely already done by the mail server,
        but not necessarily strictly.

        This check helps prevent address spoofing, but will block
        some legitimate emails.
        """

        def __init__(self, strict: bool):
            self.strict = strict

        def check(self, email: IMAPMessage) -> RuleResult:
            dkim_signatures = email.obj.get_all("DKIM-Signature")
            if not dkim_signatures:
                if self.strict:
                    return RuleResult(Action.BLOCK, "No DKIM signature")
                else:
                    return RuleResult(Action.ALLOW, None)

            try:
                result = verify_dkim_signatures(email)
            except DnsError as exc:
                logger.exception("DNS error while verifying DKIM signature")
                return RuleResult(Action.BLOCK, f"DKIM DNS error ({str(exc)})")
            except DKIMException as exc:
                logger.exception("DKIM verification error")
                return RuleResult(Action.BLOCK, f"DKIM verification error ({str(exc)})")

            if result == VerificationResult.FAIL:
                return RuleResult(Action.BLOCK, "DKIM verification failed")
            if result == VerificationResult.PASS:
                return RuleResult(Action.ALLOW, None)
            if self.strict:
                return RuleResult(Action.BLOCK, "No DKIM signature")
            else:
                return RuleResult(Action.ALLOW, None)

except ImportError:
    logger.debug("dkim extras are not installed; VerifyDKIM is not available")


RE_SMTP_MAILFROM = re.compile(r"smtp.mailfrom=(?P<mailfrom>[^;\s]+)", re.MULTILINE)


class VerifyMailFrom(Rule):
    """
    Verifies that SMTP MAIL FROM matches the From header, by
    looking for smtp.mailfrom in Authentication-Results headers.

    In strict mode, blocks emails for which smtp.mailfrom cannot be
    determined.

    This check helps prevent address spoofing, but will block
    many legitimate emails.
    """

    AUTHENTICATION_RESULTS_HEADERS = (
        "Authentication-Results",
        "ARC-Authentication-Results",
    )

    def __init__(self, strict: bool = True):
        self.strict = strict

    def check(self, email: IMAPMessage) -> RuleResult:
        smtp_mailfrom = None
        found_in_header = None
        for header in self.AUTHENTICATION_RESULTS_HEADERS:
            if m := RE_SMTP_MAILFROM.search(email.obj.get(header, "")):
                smtp_mailfrom = m.group("mailfrom")
                found_in_header = header
                break

        if smtp_mailfrom is None:
            if self.strict:
                return RuleResult(Action.BLOCK, "smtp.mailfrom not found")
            else:
                return RuleResult(Action.ALLOW, None)

        from_email = email.addr_from.email
        if smtp_mailfrom == from_email:
            return RuleResult(Action.ALLOW, None)
        else:
            return RuleResult(
                Action.BLOCK,
                f"smtp.mailfrom={smtp_mailfrom} in {found_in_header} "
                f"header does not match From={from_email}",
            )


class VerifyXMailFrom(Rule):
    """
    Verifies that the X-Mail-From header matches the From header.

    X-Mail-From is a custom header used by at least one provider
    (Fastmail) to set to the value of SMTP MAIL FROM.

    In strict mode, blocks emails which don't have the X-Mail-From header.

    This check helps prevent address spoofing, but will block
    many legitimate emails.
    """

    MAIL_FROM_HEADER = "X-Mail-From"

    def __init__(self, strict: bool = True):
        self.strict = strict

    def check(self, email: IMAPMessage) -> RuleResult:
        x_mail_from = email.obj.get(self.MAIL_FROM_HEADER)
        if self.strict and x_mail_from is None:
            return RuleResult(Action.BLOCK, f"{self.MAIL_FROM_HEADER} header is missing")
        if x_mail_from != email.addr_from.email:
            return RuleResult(
                Action.BLOCK, f"{self.MAIL_FROM_HEADER} header does not match From header"
            )
        return RuleResult(Action.ALLOW, None)


class FilterHeader(Rule):
    """
    Filter based on the value of an arbitrary email header.

    In strict mode, blocks emails which don't have the header.
    """

    def __init__(
        self,
        header: str,
        values: Iterable[str],
        mode: FilterMode,
        strict: bool = True,
    ):
        self.header = header
        self.values = set(values)
        self.mode = mode
        self.strict = strict

    def check(self, email: IMAPMessage) -> RuleResult:
        header_value = email.obj.get(self.header)
        if header_value is None:
            if self.strict:
                return RuleResult(Action.BLOCK, f"{self.header} header is missing")
            else:
                return RuleResult(Action.ALLOW, None)

        if self.mode == FilterMode.DENYLIST and header_value in self.values:
            return RuleResult(Action.BLOCK, f"{self.header} header value is in deny list")
        if self.mode == FilterMode.ALLOWLIST and header_value not in self.values:
            return RuleResult(Action.BLOCK, f"{self.header} header value is not in allow list")
        return RuleResult(Action.ALLOW, None)


class FilterFrom(Rule):
    """
    Checks if the email sender is in a list of allowed or denied addresses.
    """

    def __init__(
        self,
        mode: FilterMode,
        addresses: Iterable[str],
    ):
        self.mode = mode
        self.addresses = set()
        self.domains = set()
        for addr in addresses:
            name, domain = addr.split("@", 1)
            if name == "*":
                self.domains.add(domain)
            else:
                self.addresses.add(addr)

        logger.debug("FilterFrom: {} {} {}", mode, self.addresses, self.domains)

    def is_in_list(self, addr: str) -> bool:
        _, domain = addr.split("@", 1)
        in_list = domain in self.domains or addr in self.addresses
        return in_list

    def check(self, email: IMAPMessage) -> RuleResult:
        sender = email.addr_from.email
        if sender is None:
            return RuleResult(Action.BLOCK, "From header is missing")

        sender_in_list = self.is_in_list(sender)
        if self.mode == FilterMode.DENYLIST and sender_in_list:
            return RuleResult(Action.BLOCK, f"{sender} is in deny list")
        if self.mode == FilterMode.ALLOWLIST and not sender_in_list:
            return RuleResult(Action.BLOCK, f"{sender} is not in allow list")
        return RuleResult(Action.ALLOW, None)


class RateLimitRule(RateLimiter, Rule):
    """
    Global rate limit check.
    """

    def __init__(self, duration: datetime.timedelta, limit: int, name: str = ""):
        self.name = name
        super().__init__(duration, limit)

    def _reset(self, now: datetime.datetime) -> None:
        super()._reset(now)
        logger.trace(
            "rate limit reset {} {}/{} {}",
            self.name,
            self._limit_count,
            self.limit,
            self._limit_expiry,
        )

    def count(self) -> LimitResult:
        res = super().count()
        logger.trace(
            "rate limit {} {}/{} until {}",
            self.name,
            self._limit_count,
            self.limit,
            self._limit_expiry,
        )
        return res

    def _check(self) -> RuleResult:
        limit_result = self.count()
        if limit_result == LimitResult.EXCEEDED:
            return RuleResult(Action.BLOCK, f"rate limit {self.name} exceeded")
        return RuleResult(Action.ALLOW, None)

    def check(self, email: IMAPMessage) -> RuleResult:
        return self._check()


class RateLimitPerSenderRule(Rule):
    """
    Per-sender rate limit check.
    """

    def __init__(self, duration: datetime.timedelta, limit: int, name: str = ""):
        self.rate_limits: dict[str, RateLimitRule] = {}
        self.duration = duration
        self.limit = limit
        self._next_purge = datetime.datetime.now() + self.duration
        self.name = name

    def _purge(self, now: datetime.datetime) -> None:
        for key, rate_limit in self.rate_limits.items():
            if rate_limit._is_expired(now):
                del self.rate_limits[key]

    def _increase_and_check(self, key: str) -> RuleResult:
        now = datetime.datetime.now()

        if key in self.rate_limits:
            if self.rate_limits[key]._is_expired(now):
                self.rate_limits[key]._reset(now)
        else:
            self.rate_limits[key] = RateLimitRule(self.duration, self.limit, f"{self.name}/{key}")

        if now > self._next_purge:
            self._purge(now)
            self._next_purge = now + self.duration

        return self.rate_limits[key]._check()

    def check(self, email: IMAPMessage) -> RuleResult:
        return self._increase_and_check(email.addr_from.email)


class RateLimitPerDomainRule(RateLimitPerSenderRule):
    def check(self, email: IMAPMessage) -> RuleResult:
        _, domain = email.addr_from.email.split("@", 1)
        return self._increase_and_check(domain)


class SecurityFilter:
    """
    SecurityFilter applies the provided rules to an email to
    decide whether to block or allow it.
    """

    def __init__(self, rules: Iterable[Rule]):
        self.rules = list(rules)

    def apply(self, email: IMAPMessage) -> Action:
        for check in self.rules:
            result, reason = check.check(email)
            if result == Action.BLOCK:
                logger.log("SECURITY", "BLOCKED - {} - {}", reason, email.summary())
                return Action.BLOCK
        return Action.ALLOW


def make_security_filter(config: SecurityConfig, name_prefix: str = "") -> SecurityFilter | None:
    rules = []

    if not config.allow_from and config.allow_from_all_i_want_to_spend_it_all:
        pass
    else:
        rules.append(FilterFrom(FilterMode.ALLOWLIST, config.allow_from))

    if config.block_from:
        rules.append(FilterFrom(FilterMode.DENYLIST, config.block_from))

    if config.verify_mail_from == VerifyMode.IF_PRESENT:
        rules.append(VerifyMailFrom(strict=False))
    elif config.verify_mail_from == VerifyMode.ALWAYS:
        rules.append(VerifyMailFrom(strict=True))

    if config.verify_x_mail_from == VerifyMode.IF_PRESENT:
        rules.append(VerifyXMailFrom(strict=False))
    elif config.verify_x_mail_from == VerifyMode.ALWAYS:
        rules.append(VerifyXMailFrom(strict=True))

    if config.verify_dkim == VerifyMode.IF_PRESENT:
        assert VerifyDKIM is not None, (
            "dkim extras not installed, VerifyDKIM option is not available"
        )
        rules.append(VerifyDKIM(strict=False))
    elif config.verify_dkim == VerifyMode.ALWAYS:
        assert VerifyDKIM is not None, (
            "dkim extras not installed, VerifyDKIM option is not available"
        )
        rules.append(VerifyDKIM(strict=True))

    for fheader in config.filter_headers or []:
        if fheader.verify == VerifyMode.IF_PRESENT:
            rules.append(FilterHeader(fheader.header, fheader.values, fheader.mode, False))
        elif fheader.verify == VerifyMode.ALWAYS:
            rules.append(FilterHeader(fheader.header, fheader.values, fheader.mode, True))

    if config.rate_limit_per_sender and config.rate_limit_per_sender.limit is not None:
        rules.append(
            RateLimitPerSenderRule(
                config.rate_limit_per_sender._window_timedelta,
                config.rate_limit_per_sender.limit,
                f"{name_prefix}per-sender",
            )
        )

    if config.rate_limit_per_domain and config.rate_limit_per_domain.limit is not None:
        rules.append(
            RateLimitPerDomainRule(
                config.rate_limit_per_domain._window_timedelta,
                config.rate_limit_per_domain.limit,
                f"{name_prefix}per-domain",
            )
        )

    if config.rate_limit and config.rate_limit.limit is not None:
        rules.append(
            RateLimitRule(
                config.rate_limit._window_timedelta,
                config.rate_limit.limit,
                f"{name_prefix}global",
            )
        )

    if not rules:
        return None
    return SecurityFilter(rules)
